import random

try:
    import pyext
except:
    print "ERROR: This script must be loaded by the PD/Max pyext external"

   
# a function transposing a pitch-class
def transpose_Pitch(pitch, n):
    new_pitch = (pitch+n)%12
    return new_pitch

# a function inverting a pitch-class
def invert_Pitch(pitch, n):
    new_pitch = (n-pitch)%12
    return new_pitch

# a function analyzing transformations (inversion/transposition)
# a pitch-class set (tricord)
def analyze_set(pc_set):
    #analyzing inversion between pc0 and pc1 
    invA = (pc_set[0]+pc_set[1])%12
    #analyzing inversion between pc1 and pc2
    invB = (pc_set[1]+pc_set[2])%12
    #analyzing transposition between pc0 and pc2
    trnA = (pc_set[2]-pc_set[1])%12
    #building transformation set
    trfm = [invA, invB, trnA]
    #print('the transformation set is:', trfm)
    return trfm

# a function generating a new tricord isograph using transformation
# analysis function and pitch-class inversion function 
def generate_Isograph(pc_set,u,j,pitch):
    # new_pc_set will be isographic to pc_set
    # pitch is the "bass note" of this isograph
    new_pc_set = [pitch,0,0]
    # calling analyze_set to get the transformation set of pc_set
    transformations = analyze_set(pc_set)
    invA = transformations[0]
    invB = transformations[1]
    trnA = transformations[2]
    # using invert_pitch to generate the two other pcs
    new_pc_set[1] = invert_Pitch(pitch,(invA*u+j))
    new_pc_set[2] = invert_Pitch(new_pc_set[1],(invB*u+j))
    # returning the isograph
    return new_pc_set

# a function taking as arguments a pc_set, a list of u values, a list of j values
# and a list of "base notes" and generating a k-network, using the generate isograph
# function
def generate_Knet(pc_set, ulist, jlist, pitch_list):
    knet = [pc_set]
    for i in range(len(ulist)):
        isograph = generate_Isograph(pc_set,ulist[i],jlist[i],pitch_list[i])
        knet.append(isograph)
    return knet

# a function taking as arguments a knet, a list of u values, a list of j values
# and generating a dictionary of Hyper statements
def genHyperStatement(knet,ulist,jlist):
    ulist_complete = [1] + ulist
    jlist_complete = [1] + jlist
    dict_Hyper = {}

    for i in range(len(knet)):
        h=i+1
        for k in range(len(knet)-i):           
            l=k+h
            #print('i is:',i)
            #print('l is:',l)
            if l >= len(knet):
                continue
            elif knet[i]==knet[l]:
                continue
            elif ulist_complete[i]==ulist_complete[l]:
                f = (ulist_complete[i]*jlist_complete[l]+jlist_complete[i])%12
                print "HyperT(",f,"): ",knet[i]," ----> ",knet[l]
                key_ls1 = [tuple(knet[i]),tuple(knet[l])]
                key_tp1 = tuple(key_ls1)
                dict_Hyper[key_tp1] = ['T',f]
            else:
                f = (ulist_complete[i]*jlist_complete[l]+jlist_complete[i])%12
                print "HyperI(",f,"): ",knet[i]," ----> ",knet[l]
                key_ls2 = [tuple(knet[i]),tuple(knet[l])]
                key_tp2 = tuple(key_ls2)
                dict_Hyper[key_tp2] = ['I',f]
                #key3 = tuple(knet[l])
                #dict_Relate[key3] = [knet[l],knet[i],'I',f,'rtype2']

    return dict_Hyper

# a function taking a pc_set and the hyper statment dictionary and returns a list of two lists:
# the first is a list of pc_sets related by transposition; the second is a list of pc_sets
# related by inversion 
def select_related(pc_set,dict_Hyper):
    list_trelated = []
    list_irelated = []
    t_pc_set = tuple(pc_set)
    for key in dict_Hyper:
        if key[0] == tuple(pc_set):
            value = dict_Hyper[key]
            if value[0] == 'T':               
                list_trelated.append(list(key[1]))
            else:
                list_irelated.append(list(key[1]))
    list_related = [list_trelated,list_irelated]
    return list_related 

# a function that takes a knet and build a dict_related using
# select_related(): for each pc_set creating a list of related pc_sets
# through unidirectional T and bidirectional I, i.e., adding the return way for
# each I relations
def genDict_related(knet,dict_Hyper):
    dict_related ={}
    for pc_set in knet:
        #print('pc_set in outer loop is',pc_set)
        pc_list = []
        list_related = select_related(pc_set,dict_Hyper)
        pc_list.append(list_related[0])
        pc_list.append(list_related[1])
        #print('pc_list of', pc_set,'is',pc_list)
        dict_related[tuple(pc_set)] = pc_list
        #print('list_related[1]of', pc_set,'is',list_related[1])

    dict_value_I = {}

    for pc_set2 in knet:
        value = dict_related[tuple(pc_set2)]
        dict_value_I[tuple(pc_set2)]= value[1]
        dict_related[tuple(pc_set2)] = value[0]+value[1]

    for pc_set3 in knet:
        value = dict_value_I[tuple(pc_set3)]
        if value == []:
            continue
        else:
            for pc_set4 in value:
                #print(dict_related[tuple(pc_set3)])
                #value2 = dict_related[tuple(pc_set3)]
                #dict_related[tuple(pc_set3)] = value2[0]+value2[1]
                dict_related[tuple(pc_set4)] += [pc_set3]
                
                     
    return dict_related          

# a function taking a a start pc_set, a target pc_set, dict_related, a h value
# determining the lenth of the path and a d value determining the number or
# repetitions and returtning a path
def genPath(start,target,dict_related,h,d):
    path = [start]
    rep_count = 0
    while len(path)<h:
        value = dict_related[tuple(start)]       
        pc_set = value[random.randint(1,len(value)-1)]
        if len(path)>= 3 and pc_set == path[-3] and rep_count > d:
            continue
        elif pc_set == target:
            path.append(pc_set)
            break
        elif len(path)>= 3 and pc_set == path[-3]:
            path.append(pc_set)
            rep_count += 1
            start = pc_set
        else:
            path.append(pc_set)
            start = pc_set
    return path
    

class klump(pyext._class):
    # number of inlets and outlets
    _inlets=2
    _outlets=2

    def __init__(self):
        self.pc_set = []
        self.ulist = []
        self.jlist = []
        self.pitch_list = []
        self.knet = []
        self.dict_Hyper = {}
        self.dict_related = {}

    def list_1(self,*st):
        s = list(st)
        #print "List",s,"into first inlet"
        self.pc_set = s[2:(s[0]+2)]
        #print('pc_set is; ',self.pc_set)
        self.ulist = s[(s[0]+2):(s[0]+2+s[1])]
        #print('ulist is: ',self.ulist)
        self.jlist = s[(s[0]+2+s[1]):(s[0]+2+s[1]+s[1])]
        #print('jlist is: ',self.jlist)
        self.pitch_list = s[(s[0]+2+s[1]+s[1]):]
        #print('pitch_list is: ',self.pitch_list)
        self.knet=generate_Knet(self.pc_set, self.ulist, self.jlist, self.pitch_list)
        print "knet: ",self.knet
        print
        self.dict_Hyper = genHyperStatement(self.knet,self.ulist,self.jlist)
        print
        #print('hyper dictionary is:')
        #print(self.dict_Hyper)
        self.dict_related = genDict_related(self.knet,self.dict_Hyper)
        #print('related dictionary is:')
        #print(self.dict_related)
        for item in self.knet:
            self._outlet(1,*item)
        

    def list_2(self,*st):
        s = list(st)
        #print "List",s,"into first inlet"
        if len(s) == 2:
            self.start = self.knet[random.randint(0,(len(self.knet)-1))]
            print "start set is: ",self.start
            self.target = self.knet[random.randint(0,(len(self.knet)-1))]
            print "target set is: ",self.target
            self.h = (3 + random.randint(0,5))
            #print "h is: ",self.h
            self.d = random.randint(0,4)
            #print "d is: ",self.d
            self.path = genPath(self.start,self.target,self.dict_related,self.h,self.d)
            print "path: ",self.path
            for item in self.path:
                self._outlet(2,*item)
        elif len(s) == 8:
            self.start = s[:3]
            print "start set is: ",self.start
            self.target = s[3:6]
            print "target set is: ",self.target
            self.h = s[6:7]
            #print "h is: ",self.h
            self.d = s[7:]
            #print "d is: ",self.d
            self.path = genPath(self.start,self.target,self.dict_related,self.h,self.d)
            print "path: ",self.path
            for item in self.path:
                self._outlet(2,*item)
        elif len(s) == 5:
            self.start = s[:3]
            print "start set is: ",self.start
            self.target = self.knet[random.randint(0,(len(self.knet)-1))]
            print "target set is: ",self.target
            self.h = s[3:4]
            #print "h is: ",self.h
            self.d = s[4:]
            #print "d is: ",self.d
            self.path = genPath(self.start,self.target,self.dict_related,self.h,self.d)
            print "path: ",self.path
            for item in self.path:
                self._outlet(2,*item)
        else:
            print "to generate a path please enter a message 0 1 or"
            print "a list of two trichords a h value and a d value (8 numbers) or"
            print "a list of one trichord a h value and a d value (5 numbers)"
            
 

    def bang_2(self):
        self.path = genPath(self.start,self.target,self.dict_related,self.h,self.d)
        print "path: ",self.path
        for item in self.path:
            self._outlet(2,*item)
        

    


